/* FILE: driver.cc                 -*-Mode: c++-*-
 *
 * Abstract driver class; fills states and steps evolvers.
 * Also registers the command "Oxs_Run" with the Tcl intepreter.
 *
 */

#include "oc.h"
#include "nb.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "evolver.h"
#include "key.h"
#include "energy.h"	// Needed to make MSVC++ 5 happy

#include <string>

/* End includes */

// Constructor
Oxs_Driver::Oxs_Driver
( const char* name,        // Child instance id
  Oxs_Director* newdtr,    // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr       // Args; should include an 
                          /// "evolver <evolver-name>" pair.
  ) : Oxs_Ext(name,newdtr,safe_interp,argstr)
{
  // Get evolver name specification
  CheckInitValueParamCount("evolver",1);
  const string& evolver_name = (*FindInitValue("evolver"))[0];

  // Put dependency lock on specified evolver
  Oxs_Ext* tempobj = director->FindExtObject(evolver_name);
  evolver_key.Set(dynamic_cast<Oxs_Evolver*>(tempobj));
  if(evolver_key.GetPtr()==NULL) {
    string msg="Unable to identify Oxs_Evolver object with name "
      + evolver_name;
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("evolver");
  // evolver holds a dependency lock on the Oxs_Evolver
  // that won't be freed until *this is destroyed.

  // Reserve state space in director
  director->ReserveSimulationStateRequest(2);

}

//Destructor
Oxs_Driver::~Oxs_Driver()
{}

// Oxs_Driver version of Init().  All children of Oxs_Driver *must*
// call this function in their Init() routines.  The main purpose
// of this function is to initialize the current state.
BOOL Oxs_Driver::Init_Driver()
{
  // Initialize current state from initial state provided by
  // concrete child class.
  current_state = GetInitialState();
  return (current_state.GetPtr() != NULL);
}

int Oxs_Driver::Run(string& result,int /*argc*/, char** /*argv*/,
	Tcl_Interp *interp)
{ // Returns 0 on success.  When successful, stores in result a
  // Tcl script to be evaluated by the caller.  When not successful,
  // result contains an error message.
  result.erase();

  if(current_state.GetPtr() == NULL) {
    // Current state is not initialized.
    string msg="Current state in Oxs_Driver is not initialized;"
      " This is probably the fault of the child class "
      + string(ClassName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  if (Done(current_state.GetReadReference(), result)) {
    return 0;
  }

  // Repeat forever, until Done() or Return() breaks us out
  Oxs_Key<Oxs_Evolver> temp_key = evolver_key;
  Oxs_Evolver& evolver = temp_key.GetWriteReference();
  while (1) {

    // Attempt to take a step.
    Oxs_Key<Oxs_SimState> next_state;
    director->GetNewSimulationState(next_state);
    FillStatePartA(current_state.GetReadReference(),
                   next_state.GetWriteReference());
    if( evolver.Step(*this,current_state,next_state) ) {
      // Good step.  Release read lock on old current_state,
      // and copy key from next_state.
      current_state = next_state;
  
      // Give interface a chance to know we have a new state.
      NewState(current_state.GetReadReference(), result);
    }
  
    if (StageDone(current_state.GetReadReference(), result)) {
      // We're about to change state, so evaluate any result
      // script now before that happens.
      // NOTE: a rewrite of this routine should push this evaluation
      // up to the caller.
      char* buf = new char[result.length()+1];
      strcpy(buf,result.c_str());
      int code = Tcl_Eval(interp, buf);
      delete[] buf;
      if (code) {
        result = Tcl_GetStringResult(interp);
	Tcl_ResetResult(interp);
	return 1;
      } else {
	// Script completed successfully.  Don't run it again.
	// However, we want to return to break out of this
	// while loop and return to Tcl's event loop, so place
	// in result a script that does nothing to force that return.
	result = "\n";
      }

      // Setup for next stage
      director->GetNewSimulationState(next_state);
      Oxs_SimState& nstate = next_state.GetWriteReference();
      nstate = current_state.GetReadReference(); // Copy data
      nstate.stage_number++;
      nstate.stage_iteration_count=0;
      // Spins are the same, so keep the same iteration_count. (???)
      next_state.GetReadReference();  // Release write lock
      current_state = next_state;
    }

    // Quit if the problem is solved.
    if (Done(current_state.GetReadReference(), result)) {
      return 0;
    }

    // If a script has accumulated, return so it can be evaluated.
    if (result.length()) {
	return 0;
    }
  }

}
