/* FILE: randomvectorfieldinit.h      -*-Mode: c++-*-
 *
 * Random vector field initializer, derived from Oxs_VectorFieldInit class.
 *
 */

#include "oc.h"
#include "nb.h"

#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "planerandomvectorfieldinit.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_PlaneRandomVectorFieldInit);

/* End includes */


// Constructor
Oxs_PlaneRandomVectorFieldInit::Oxs_PlaneRandomVectorFieldInit(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorFieldInit(name,newdtr,safe_interp,argstr)
{
  // Process arguments.
  min_norm=GetRealInitValue("min_norm");
  max_norm=GetRealInitValue("max_norm");

  CheckInitValueParamCount("plane_normal",3);
  const vector<string>& vstr = *FindInitValue("plane_normal");
  plane_normal.x=Nb_Atof(vstr[0].c_str());
  plane_normal.y=Nb_Atof(vstr[1].c_str());
  plane_normal.z=Nb_Atof(vstr[2].c_str());
  plane_normal.SetMag(1.0);
  DeleteInitValue("plane_normal");

  VerifyAllInitArgsUsed();
}

BOOL Oxs_PlaneRandomVectorFieldInit::Init()
{
  return 1;
}

void
Oxs_PlaneRandomVectorFieldInit::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<ThreeVector>& array) const
{
  const REAL8m norm_range=max_norm-min_norm;
  REAL8m dot=0.0;
  ThreeVector vec;
  UINT4m i,size=mesh->Size();
  array.AdjustSize(mesh);
  for(i=0;i<size;i++) {
    // First produce vector perpendicular to plane
    int j=0;
    while(1) {
      vec.Random(1.0);
      dot = plane_normal*vec;
      if(fabs(dot)<0.99) break;
      if(++j>32) {
	string msg=
	  string("Programming error?  Unable to obtain vector"
		 " in specified plane in"
		 " Oxs_PlaneRandomVectorFieldInit::FillMeshValue()"
		 " method of instance ")
	  + string(InstanceName());
	throw Oxs_Ext::Error(msg.c_str());
      }
    }
    vec -= dot*plane_normal;

    // Adjust to proper size
    REAL8m mag = min_norm;
    if(norm_range!=0.0) {
      // Vary magnitude
      mag += norm_range*Oc_UnifRand();
    }
    vec.SetMag(mag);
    array[i] = vec;
  }
}

void
Oxs_PlaneRandomVectorFieldInit::FillMeshValueOrthogonal
(const Oxs_Mesh* mesh,
 const Oxs_MeshValue<ThreeVector>& orth_array,
 Oxs_MeshValue<ThreeVector>& array) const
{
  if(!orth_array.CheckMesh(mesh)) {
    string msg=
      string("Mesh size incompatibility detected in"
	     " Oxs_PlaneRandomVectorFieldInit::FillMeshValueOrthogonal()"
	     " method of instance ")
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  const REAL8m norm_range=max_norm-min_norm;
  REAL8m dot=0.0;
  ThreeVector vec;
  UINT4m i,size=mesh->Size();
  array.AdjustSize(mesh);
  for(i=0;i<size;i++) {
    // We need a vector that is orthogonal to both plane and
    // orth_array[i].  First try 'plane x orth_array[i]'
    vec = orth_array[i];
    vec.SetMag(1.0);
    vec ^= plane_normal;
    if(vec.MagSq()>0.0) {
      // Randomly select +/-
      if(Oc_UnifRand()<0.5) vec *= -1;
    } else {
      // orth_array[i] is either 0 or parallel to plane,
      // so just produce a vector orthogonal to plane from
      // scratch.
      int j=0;
      while(1) {
	vec.Random(1.0);
	dot = plane_normal*vec;
	if(fabs(dot)<0.99) break;
	if(++j>32) {
	  string msg=
	    string("Programming error?  Unable to obtain vector"
		   " in specified plane in"
		   " Oxs_PlaneRandomVectorFieldInit::FillMeshValueOrthogonal()"
		   " method of instance ")
	    + string(InstanceName());
	  throw Oxs_Ext::Error(msg.c_str());
	}
      }
      vec -= dot*plane_normal;
    }

    // Adjust to proper size
    REAL8m mag = min_norm;
    if(norm_range!=0.0) {
      // Vary magnitude
      mag += norm_range*Oc_UnifRand();
    }
    vec.SetMag(mag);
    array[i] = vec;
  }
}
