/* FILE: scriptscalarfieldinit.h      -*-Mode: c++-*-
 *
 * Tcl script scalar field initializer, derived from Oxs_ScalarFieldInit
 * class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"

#include "util.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "scriptscalarfieldinit.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_ScriptScalarFieldInit);

/* End includes */

// Constructor
Oxs_ScriptScalarFieldInit::Oxs_ScriptScalarFieldInit(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp_, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_ScalarFieldInit(name,newdtr,safe_interp_,argstr),
    safe_interp(safe_interp_)
{
  // Process arguments.
  const vector<string>* scriptlistptr = FindInitValue("script");
  if(scriptlistptr==NULL) {
    string msg="No script specified in input block for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  // Copy script into instance memory
  scriptlist.reserve(scriptlistptr->size());
  for(UINT4m iarg=0;iarg<scriptlistptr->size();++iarg) {
    scriptlist.push_back((*scriptlistptr)[iarg]);
  }
  DeleteInitValue("script");

  VerifyAllInitArgsUsed();
}

BOOL Oxs_ScriptScalarFieldInit::Init()
{ return 1; }

void 
Oxs_ScriptScalarFieldInit::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<REAL8m>& array) const
{
  // Fill array with values from import Tcl script.
  UINT4m i;
  UINT4m size=mesh->Size();
  Oxs_Box bbox;
  mesh->GetBoundingBox(bbox);
  ThreeVector loc;

  array.AdjustSize(mesh);

#if (TCL_MAJOR_VERSION < 8 || (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==0))
  // For Tcl eval string interface
  string script=scriptlist[0];
  for(UINT4m iarg=1;iarg<scriptlist.size();iarg++) {
    script += " ";
    script += scriptlist[iarg];
  }
  char extents[1024];
  Oc_Snprintf(extents,sizeof(extents),
              "%.16g %.16g %.16g %.16g %.16g %.16g",
              bbox.GetMinX(),bbox.GetMinY(),bbox.GetMinZ(),
              bbox.GetMaxX(),bbox.GetMaxY(),bbox.GetMaxZ());
  char cmdbuf[4096];
  Oxs_SplitList veclist;
  for(i=0;i<size;i++) {
    mesh->Center(i,loc);
    Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
                "%s %.16g %.16g %.16g %s",
                script.c_str(),loc.x,loc.y,loc.z,extents);
    if(Tcl_Eval(safe_interp,cmdbuf)!=TCL_OK) {
      string msg=string("Error detected during construction of ")
        + string(InstanceName()) + string("; Tcl script error: ")
        + string(Tcl_GetStringResult(safe_interp));
      throw Oxs_Ext::Error(msg.c_str());
    }
    veclist.Split(Tcl_GetStringResult(safe_interp));
    if(veclist.Count()!=1) {
      string msg
        = string("Error detected during construction of ")
        + string(InstanceName())
        + string(" Return script value is not a single scalar: ")
        + string(Tcl_GetStringResult(safe_interp));
      throw Oxs_Ext::Error(msg.c_str());
    }
    array[i]=Nb_Atof(veclist[0]);
  }
#else // ! (Tcl version < 8.1)
  // For Tcl eval obj interface
  // NOTE: We should wrap this up inside a C++ class object
  // to get proper cleanup if an exception is thrown.  For
  // now, we'll just live with the MEMORY LEAK.
  UINT4m objcmd_argc=static_cast<UINT4m>(scriptlist.size())+9;
  Tcl_Obj** objcmd = new Tcl_Obj*[objcmd_argc];
  for(UINT4m iarg=0;iarg<scriptlist.size();iarg++) {
    objcmd[iarg] 
      = Tcl_NewStringObj(Oc_AutoBuf(scriptlist[iarg].c_str()),
			 static_cast<int>(scriptlist[iarg].size()));
  }
  objcmd[objcmd_argc-9] = Tcl_NewDoubleObj(0.); // Dummy value
  objcmd[objcmd_argc-8] = Tcl_NewDoubleObj(0.); // ""
  objcmd[objcmd_argc-7] = Tcl_NewDoubleObj(0.); // ""
  objcmd[objcmd_argc-6] = Tcl_NewDoubleObj(bbox.GetMinX());
  objcmd[objcmd_argc-5] = Tcl_NewDoubleObj(bbox.GetMinY());
  objcmd[objcmd_argc-4] = Tcl_NewDoubleObj(bbox.GetMinZ());
  objcmd[objcmd_argc-3] = Tcl_NewDoubleObj(bbox.GetMaxX());
  objcmd[objcmd_argc-2] = Tcl_NewDoubleObj(bbox.GetMaxY());
  objcmd[objcmd_argc-1] = Tcl_NewDoubleObj(bbox.GetMaxZ());
  for(i=0;i<objcmd_argc;i++) Tcl_IncrRefCount(objcmd[i]);
  for(i=0;i<size;i++) {
    mesh->Center(i,loc);
    Tcl_SetDoubleObj(objcmd[objcmd_argc-9],loc.x);
    Tcl_SetDoubleObj(objcmd[objcmd_argc-8],loc.y);
    Tcl_SetDoubleObj(objcmd[objcmd_argc-7],loc.z);
    if(Tcl_EvalObjv(safe_interp,objcmd_argc,objcmd,TCL_EVAL_GLOBAL)
       !=TCL_OK) {
      for(i=0;i<objcmd_argc;i++) Tcl_DecrRefCount(objcmd[i]);
      delete[] objcmd;
      string msg=string("Error detected during construction of ")
        + string(InstanceName()) + string("; Tcl script error: ")
        + string(Tcl_GetStringResult(safe_interp));
      throw Oxs_Ext::Error(msg.c_str());
    }
    int objc;
    double result;
    Tcl_Obj** objv;
    if(Tcl_ListObjGetElements(NULL,Tcl_GetObjResult(safe_interp),
			      &objc,&objv) != TCL_OK
       || objc != 1
       || Tcl_GetDoubleFromObj(NULL,objv[0],&(result)) != TCL_OK
       ) {
      for(i=0;i<objcmd_argc;i++) Tcl_DecrRefCount(objcmd[i]);
      delete[] objcmd;
      string msg
        = string("Error detected during construction of ")
        + string(InstanceName())
        + string(" Return script value is not a single scalar: ")
        + string(Tcl_GetStringResult(safe_interp));
      throw Oxs_Ext::Error(msg.c_str());
    }
    array[i]=result;
  }

  if(objcmd!=NULL) {
    for(i=0;i<objcmd_argc;i++) Tcl_DecrRefCount(objcmd[i]);
    delete[] objcmd;
  }
#endif // (Tcl version < 8.1)
}

