#!/bin/sh
#    v--Edit here if necessary \
exec tclsh "$0" ${1+"$@"}

set progvers "MIF Conversion utility 1.2.0.1"

if {[llength $argv]!=2} {
    puts stderr "Usage: tclsh mifconvert.tcl <infile> <outfile> "
    exit 1
}

global tcl_precision
set tcl_precision 17

set PI [expr {4*atan(1.0)}]
set MU0 [expr {4*$PI*1e-7}]

# ExtractRecord finds the first record in the import list with id $rid.
# The record (both id and params) are removed from the list.  The
# return value is the param field for the record.  An error is thrown
# if no record id matches $rid.
proc ExtractRecord { listname rid } {
    upvar $listname inlist
    set index [lsearch -exact $inlist $rid]
    if {$index<0} {
	error "Record id $rid not found in input list"
    }
    set params [lindex $inlist [expr {$index + 1}]]
    set inlist [lreplace $inlist $index [expr {$index + 1}]]
    return $params
}

# EatAllRecords is similar to ExtractRecord, except that if multiple
# record id matches are found, then all are removed and the param field
# for the last is returned.
proc EatAllRecords { listname rid } {
    upvar $listname inlist
    set params [ExtractRecord inlist $rid]
    while {![catch {ExtractRecord $listname $rid} tmp]} {
	set params $tmp
    }
    return $params
}

# Open input file
set infilename [lindex $argv 0]
set infile [open $infilename "r"]

# Check file type
set ws "\[ \t\n\r\]"   ;# White space
set bs "\[^ \t\n\r\]"  ;# Black space
set major 0
set minor 0
if {![regexp -- "^#$ws*MIF$ws*(\[0-9\])\\.(\[0-9\])$ws*$" [gets $infile] \
	dummy major minor] || $major!=1 || $minor!=1} {
    puts stderr "Input file $infilename not in proper format"
    exit 2
}

set workline {}
while {[gets $infile line]>=0} {
    if {[regexp -- {^(.*)\$} $line dummy leader]} {
	# Continued line
	append workline $leader
	continue
    }
    append workline $line

    # Skip comment and blank lines
    if {[string match {} $workline] || \
	    [regexp -- "^$ws*#" $workline]} {
	set workline {}
	continue
    }

    # Otherwise parse into record id + parameter list
    if {![regexp -- {^([^:=]+)[:=]+(.+)$} $workline dummy rid params]} {
	puts stderr "Invalid input line: $workline"
	exit 3
    }

    # Strip whitespace and case from record id
    regsub -all -- $ws $rid {} rid
    set rid [string tolower $rid]

    # Strip comments from parameter list
    set params [string trim $params]
    set templist [split $params "#"]
    set tempparams {}
    foreach piece $templist {
	append tempparams $piece
	if {![catch {llength $tempparams} size] && $size>0} {
	    break ;# Complete list, so pitch remainder (which is comment)
	}
	append tempparams "#"
    }
    if {[catch {llength $tempparams} size] || $size<1} {
	puts stderr "Record $rid has invalid parameter list: $params"
	exit 4
    }
    set params $tempparams

    # Store record id + parameter list in input list
    lappend inlist $rid $params

    set workline {}
}
close $infile

##############################################################
# Write header
set outfilename [lindex $argv 1]
set outfile [open $outfilename "w"]
puts $outfile "# MIF 2.0"
puts $outfile "# This file converted from MIF 1.1 format by $progvers"

##############################################################
# Write mesh info
set xspan [EatAllRecords inlist "partwidth"]
set yspan [EatAllRecords inlist "partheight"]
set zspan [EatAllRecords inlist "partthickness"]
set cellsize [EatAllRecords inlist "cellsize"]
puts $outfile "Specify Oxs_SectionAtlas:atlas {
  world { Oxs_RectangularSection {
    xrange {0 $xspan}
    yrange {0 $yspan}
    zrange {0 $zspan}
  }  }
}

Specify Oxs_RectangularMesh:mesh {
  cellsize {$cellsize $cellsize $zspan}
  atlas Oxs_SectionAtlas:atlas
}"

##############################################################
# Write crystalline anisotropy info 
set anisk1 [EatAllRecords inlist "k1"]
set anistype [EatAllRecords inlist "anisotropytype"]
if {[catch {EatAllRecords inlist "anisotropydir1"} anisdir1]} {
    set anisdir1 "1 0 0"
}
if {[catch {EatAllRecords inlist "anisotropydir2"} anisdir2]} {
    set anisdir2 "0 1 0"
}
if {[catch {EatAllRecords inlist "anisotropyinit"} anisinit]} {
    set anisinit "Constant"
}
set anisinit [string tolower $anisinit]
if {[string match uniformxy $anisinit]} {
    set axisinit_fmt {
	Oxs_PlaneRandomVectorFieldInit {
	    min_norm 1.0
	    max_norm 1.0
	    plane_normal {0 0 1}
	}
    }
} elseif {[string match uniforms2 $anisinit]} {
    set axisinit_fmt {
	Oxs_RandomVectorFieldInit {
	    min_norm 1.0
	    max_norm 1.0
	}
    }
} else {
    # Assume anisinit == "constant"
    set axisinit_fmt {
	Oxs_UniformVectorFieldInit {
	    norm 1
	    vector {%s}
	}
    }
}

if {[string match cubic $anistype]} {
    # Cubic anisotropy
    set axis1init_str [format $axisinit_fmt $anisdir1]
    set axis2init_str [format $axisinit_fmt $anisdir2]
puts $outfile "
Specify Oxs_CubicAnisotropy {
    K1  { Oxs_UniformScalarFieldInit { value $anisk1 } }
    axis1 { $axis1init_str }
    axis2 { $axis2init_str }
}"
} else {
    # Uniaxial anisotropy
    set axisinit_str [format $axisinit_fmt $anisdir1]
puts $outfile "
Specify Oxs_UniaxialAnisotropy {
    K1  { Oxs_UniformScalarFieldInit { value $anisk1 } }
    axis { $axisinit_str }
}"
}

##############################################################
# Write exchange info
set A [EatAllRecords inlist "a"]
puts $outfile "
Specify Oxs_UniformExchange {
  A  $A
}"

##############################################################
# Write demag info
set demag_type [EatAllRecords inlist "demagtype"]
if {![string match "none" $demag_type]} {
    if {[string match "fastpipe" $demag_type]} {
	puts stderr "ERROR: 2D demag request, \"fastpipe\", not\
		supported.\nABORT"
	puts $outfile "error {INCOMPLETE MIF FILE}"
	exit 5
    }
    if {![string match "constmag" $demag_type]} {
	puts stderr "WARNING: Requested demag type, $demag_type,\
		not supported.  Using constmag instead."
    }
    puts $outfile "
Specify Oxs_Demag {}"
}


##############################################################
# Write applied field info.
# Field Range line in MIF 1.1 has format
#     x0 y0 z0  x1 y1 z1  steps  (+optional control point specs)
if {![catch {EatAllRecords inlist "fieldtype"} Htype]} {
    if {![string match -nocase "uniform" $Htype]} {
	puts stderr \
	   "ERROR: Unsupported applied field type: $Htype"
	puts stderr "ABORT"
	puts $outfile "error {INCOMPLETE MIF FILE}"
	exit 10
    }
}

set rangecount 0
set stagecount 1
set lastfield {0 0 0}
while {![catch {ExtractRecord inlist "fieldrange"} range]} {
    if {$rangecount==0} {
	puts $outfile "\nSpecify Oxs_UZeeman \{\n   Hrange \{"
	set stagecount 0
    }
    incr rangecount
    if {[llength $range]>7} {
	puts stderr "WARNING: Field Range control point specs not supported."
	puts stderr "  Ignoring spec request \"[lrange $range 7 end]\" on\
		field range line $rangecount."
	set range [lrange $range 0 6]
    }
    # Convert units from T to A/m
    set newrange {}
    foreach comp [lrange $range 0 5] {
	lappend newrange [expr {$comp/$MU0}] 
    }
    lappend newrange [lindex $range 6]
    puts $outfile "      [list $newrange]"
    if {[lindex $newrange 6]==0} {
	incr stagecount
	set lastfield [lrange $newrange 0 2]
    } else {
	if {$stagecount==0 || \
		[lindex $lastfield 0]!=[lindex $newrange 0] || \
		[lindex $lastfield 1]!=[lindex $newrange 1] || \
		[lindex $lastfield 2]!=[lindex $newrange 2]} {
	    incr stagecount
	}
	incr stagecount [lindex $newrange 6]
	set lastfield [lrange $newrange 3 5]
    }
}
if {$rangecount>0} {
    puts $outfile "   \}\n\}"
}

##############################################################
# Write evolver info
if {[catch {EatAllRecords inlist "doprecess"} do_precess]} {
    set do_precess 1
}
if {[catch {EatAllRecords inlist "gyratio"} gamma]} {
    set gamma 2.21e5
}
if {[catch {EatAllRecords inlist "dampcoef"} alpha]} {
    set alpha 0.5
}

puts $outfile "
Specify Oxs_EulerEvolve {
  do_precess $do_precess
  gamma $gamma
  alpha $alpha
}"


##############################################################
# Write driver info
set Ms [EatAllRecords inlist "ms"]
if {[catch {EatAllRecords inlist "convergemxhvalue"} stop_mxh] && \
    [catch {EatAllRecords inlist "converge|mxh|value"} stop_mxh]} {
    set stop_mxh 1e-5
}
if {![catch {EatAllRecords inlist "defaultcontrolpointspec"} stop_spec]} {
    set stop_spec [string tolower [string trim $stop_spec]]
    if {![regexp "^-torque$ws+($bs+)$" $stop_spec \
	    dummy stop_mxh]} {
	puts stderr \
		"WARNING: Unsupported Default Control Point Spec: $stop_spec"
	puts stderr "  Using \"-torque $stop_mxh\" instead."
    }
}
set stopping_dm_dt [expr {$stop_mxh*$Ms*$gamma*sqrt(1+$alpha*$alpha)}]
set stopping_dm_dt [expr {$stopping_dm_dt*180e-9/$PI}] ;# Convert to deg/ns
if {[catch {EatAllRecords inlist "maxtimestep"} max_timestep]} {
    set max_timestep 10e-9
}
if {[catch {EatAllRecords inlist "mintimestep"} min_timestep]} {
    set min_timestep 1e-18
}
if {[catch {EatAllRecords inlist "partshape"} part_shape]} {
    set part_shape rectangle
}
if {[string match "rectangle" $part_shape]} {
    set Ms_str "Oxs_UniformScalarFieldInit { value $Ms }"
} elseif {[string match "ellipse" $part_shape]} {
    puts $outfile {
proc Ellipse { Ms x y z xmin ymin zmin xmax ymax zmax } {
    set xcenter [expr ($xmax-$xmin)/2.]
    set ycenter [expr ($ymax-$ymin)/2.]
    set xrad [expr $x/$xcenter -1 ]
    set yrad [expr $y/$ycenter -1 ]
    set test [expr $xrad*$xrad+$yrad*$yrad]
    if {$test>1.0} {return 0}
    return $Ms
}}
    set Ms_str "Oxs_ScriptScalarFieldInit { script {Ellipse $Ms} }"
} elseif {[string match "ellipsoid" $part_shape]} {
    puts $outfile {
proc Ellipsoid { Ms x y z xmin ymin zmin xmax ymax zmax } {
    set xcenter [expr ($xmax-$xmin)/2.]
    set ycenter [expr ($ymax-$ymin)/2.]
    set zcenter [expr ($zmax-$zmin)/2.]
    set xrad [expr $x/$xcenter -1 ]
    set yrad [expr $y/$ycenter -1 ]
    set zrad [expr $z/$zcenter -1 ]
    set test [expr $xrad*$xrad+$yrad*$yrad+$zrad*$zrad]
    if {$test>1.0} {return 0}
    return $Ms
}}
    set Ms_str "Oxs_ScriptScalarFieldInit { script {Ellipsoid $Ms} }"
} else {
    puts stderr "ERROR: Unsupported part shape: $part_shape"
    puts stderr "ABORT"
    puts $outfile "error {INCOMPLETE MIF FILE}"
    exit 6
}
if {[catch {EatAllRecords inlist "initmag"} init_mag]} {
    set init_mag "random"
}
set init_name [string tolower [lindex $init_mag 0]]
switch -exact -- $init_name {
    "random" {
	set m0_str "Oxs_RandomVectorFieldInit { min_norm 1  max_norm 1 }"
    }
    "uniform" {
	if {[llength $init_mag]!=3} {
	    puts stderr "ERROR: Invalid mag init record: $init_mag"
	    puts stderr "ABORT"
	    puts $outfile "error {INCOMPLETE MIF FILE}"
	    exit 7
	}
	set theta [expr [lindex $init_mag 1]*$PI/180.]
	set phi [expr [lindex $init_mag 2]*$PI/180.]
	set x [expr {sin($theta)*cos($phi)}]
	set y [expr {sin($theta)*sin($phi)}]
	set z [expr {cos($theta)}]
	set m0_str "Oxs_UniformVectorFieldInit { norm 1  vector {$x $y $z} }"
    }
    "vortex" {
	puts $outfile {
proc Vortex { x y z xmin ymin zmin xmax ymax zmax } {
    set xcenter [expr ($xmax-$xmin)/2.]
    set ycenter [expr ($ymax-$ymin)/2.]
    set xrad [expr $x-$xcenter]
    set yrad [expr $y-$ycenter]
    set normsq [expr $xrad*$xrad+$yrad*$yrad]
    if {$normsq <= $xcenter*$ycenter*0.05} {return "0 0 1"}
    return [list [expr -1*$yrad] $xrad 0]
}}
        set m0_str "Oxs_ScriptVectorFieldInit { script Vortex norm  1 }"
    }
    "avffile" {
	if {[llength $init_mag]!=2} {
	    puts stderr "ERROR: Invalid mag init record: $init_mag"
	    puts stderr "ABORT"
	    puts $outfile "error {INCOMPLETE MIF FILE}"
	    exit 8
	}
	set filename [lindex $init_mag 1]
        set m0_str "Oxs_FileVectorFieldInit \
		{[list norm 1 file $filename]}"
    }
    default {
	puts stderr "ERROR: Unsupported mag init: $init_mag"
	puts stderr "ABORT"
	puts $outfile "error {INCOMPLETE MIF FILE}"
	exit 9
    }
}

puts $outfile "
Specify Oxs_StandardDriver {
 evolver Oxs_EulerEvolve
 mesh :mesh
 stage_iteration_limit 0
 total_iteration_limit 0
 number_of_stages $stagecount
 stopping_dm_dt $stopping_dm_dt
 min_timestep $min_timestep
 max_timestep $max_timestep
 Ms { $Ms_str }
 m0 { $m0_str }
}"


##############################################################
# Write Miscellaneous block
set basename [EatAllRecords inlist "baseoutputfilename"]
puts $outfile "
Miscellaneous {
   basename [list $basename]
}"

##############################################################
# Dump unused fields
puts $outfile "\n
##########################################
# Unused fields:
proc SKIP { args } {}
SKIP {"
foreach {id val} $inlist { puts $outfile "  $id: $val" }
puts $outfile "}"

close $outfile
puts "TRANSLATED [list $infilename] -> [list $outfilename]"
puts " Please check [list $outfilename] for correctness and additional edits."
