# FILE: server.tcl
#
# A generic server that can be instantiated and customized.
#

Oc_Class Net_Server {

    const public common defaultAllowList {}
    const public common defaultDenyList {}
    const public common forbidRemoteConnections 1

  # Protocol to be used on connections established by this server
  # Need good check that doesn't depend on naming conventions
  const public variable protocol 

  # The alias by which this server identifies itself
  const public variable alias = {}

  # Boolean - does this server register with a services directory?
  const public variable register = 1

  # List of remote hostname patterns allowed to make connections.
  public variable allow {
        Oc_Log Log "$this: New allow: $allow" status $class
  }

  # List of remote hostname patterns forbidden to make connections.
  public variable deny {
        Oc_Log Log "$this: New deny: $deny" status $class
  }

  # Track whether the server instance owns, and must clean up
  # the protocol instances it uses.
  const private variable imadeprotocol = 0

  # The socket through which this server accepts connections
  private variable socket

  private variable port

  Constructor {args} {
    set allow $defaultAllowList
    set deny $defaultDenyList
    eval $this Configure $args
    if {![info exists protocol]} {
      # Provide a default protocol
      Net_Protocol New protocol -name "Default protocol"
      set imadeprotocol 1
    }
    # Die when $protocol dies.
    Oc_EventHandler New _ $protocol Delete [list $this Delete] \
        -groups [list $this]
  }

  # Start the server accepting connections port
  method Start { _port } { 
    if {[info exists socket]} {
      error "Server already started"
    }
    if {$forbidRemoteConnections} {
        set socket [socket -server [list $this VerifyHostAccess] \
            -myaddr localhost $_port]
    } else {
        set socket [socket -server [list $this VerifyHostAccess] \
            $_port]
    }
    set port [lindex [fconfigure $socket -sockname] 2]
    Oc_Log Log "Server $this started on port $port..." status $class
    if {$register} {
        $this Register
    }
  }

  method VerifyHostAccess {accsock clientipaddr clientport} {
    set myinfo [fconfigure $accsock -sockname]
    set remoteinfo [fconfigure $accsock -peername]
    set myhostname [lindex $myinfo 1]
    set remotehostname [lindex $remoteinfo 1]
    if {![string match $myhostname $remotehostname]} {
      if {[string match $clientipaddr $remotehostname]} {
        $this RefuseConnection $accsock $remotehostname $clientport
        return
      }
      set allowconnection 0
      foreach pattern $allow {
        if {[string match $pattern $remotehostname]} {
          set allowconnection 1
          break
        }
      }
      foreach pattern $deny {
        if {[string match $pattern $remotehostname]} {
          set allowconnection 0
          break
        }
      }
      if {!$allowconnection} {
        $this RefuseConnection $accsock $remotehostname $clientport
        return
      }
    }
    Net_Connection New _ -socket $accsock -protocol $protocol
  }

  private method RefuseConnection {accsock rhost cport} {
    puts $accsock "tell Connection refused"
    close $accsock
    Oc_Log Log "Server $this refused connection from $rhost:$cport" \
	    status $class
  }

  # Returns true if server is accepting connections, false otherwise.
  method IsListening {} {
    return [info exists socket]
  }

  # Return port we're listening on
  method Port {} {
      if {[info exists port]} {
          return $port
      }
      error "server is not listening"
  }

  # Stop accepting connections
  method Stop {} {
    if {![info exists socket]} {
      error "Server not started"
    }
    Oc_Log Log "Stopping server $this ..." status $class
    if {$register} {
        set socketinfo [fconfigure $socket -sockname]
        set port [lindex $socketinfo 2]
        $this Deregister
    }
    unset port
    close $socket
    unset socket
  }

    private variable account

    private method Register {} {
        if {[catch {Net_Account New account -hostname localhost} msg]} {
            catch {unset account}
            $this RegisterFail $msg {}
            return
        }
        # Current implementation requires a persistent connection to the
        # account server through a Net_Account instance.  Consider whether
        # this can be relaxed after a successful registration?
        Oc_EventHandler New _ $account Delete \
                [list $this RegisterFail "account died" {}] -groups [list $this]
        if {[$account Ready]} {
            $this AccountReady
        } else {
            Oc_EventHandler New _ $account Ready [list $this AccountReady] \
                -oneshot 1 -groups [list $this]
        }
    }

    method AccountReady {} {
	regexp _${class}(.*) $this -> mySerial
        set qid [$account Send register [pid]:$mySerial $alias $port]
        Oc_EventHandler New _ $account Reply$qid \
	     [list $this RegisterReply $qid] \
             -groups [list $this $this-$qid]
        Oc_EventHandler New _ $account Timeout$qid \
             [list $this RegisterFail "timeout on registration reply" $qid] \
             -groups [list $this $this-$qid]
    }

    method RegisterReply {qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set reply [$account Get]
        switch -- [lindex $reply 0] {
            0 {
                Oc_Log Log "Server $this registered successfully" status $class
            }
            default {
                $this RegisterFail "registration error:\
			[lrange $reply 1 end]" $qid
            }
        }
    }

    method RegisterFail {msg qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set accountname [Net_Account DefaultAccountName]
        $this Delete
        global errorInfo
        set errorInfo "No stack available"
        Oc_Log Log "Unable to register server with\
                localhost:$accountname:\n\t$msg" warning $class
    }

    private method Deregister {} {
        if {![info exists account] || ![$account Ready]} {
            return
        }
	regexp _${class}(.*) $this -> mySerial
        set qid [$account Send deregister [pid]:$mySerial $port]
        Oc_EventHandler New _ $account Reply$qid \
                [list $this DeregisterReply $qid] \
                -groups [list $this $this-$qid]
        Oc_EventHandler New _ $account Timeout$qid \
                [list $this DeregisterFail "timeout on dereg. reply" $qid]] \
                -groups [list $this $this-$qid]
    }

    method DeregisterReply {qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set reply [$account Get]
        switch -- [lindex $reply 0] {
            0 {
                Oc_Log Log "Server $this deregistered successfully"\
			status $class
            }
            default {
                $this DeregisterFail "deregistration error" $qid
            }
        }
    }

    method DeregisterFail {msg qid} {
        Oc_EventHandler DeleteGroup $this-$qid
        set accountname [Net_Account DefaultAccountName]
        # Ought to send up fireworks; this is bad.
        $this Delete
        global errorInfo
        set errorInfo "No stack available"
        Oc_Log Log "Unable to deregister server with\
                localhost:$accountname:\n\t$msg" warning $class
    }

  Destructor {
    if {[info exists socket]} {
        # Stop method launches Deregister in background.  Should we wait
        # to confirm it before dying?   Proper usage is to Stop server,
        # then delete it.  More thought needed here...
        $this Stop
    }
    Oc_EventHandler DeleteGroup $this
    if {$imadeprotocol} {
        $protocol Delete
    }
  }

}
